//--------------------------------------------------------------
// File     : xmc2go_ub_nokia5110.c
// Datum    : 16.05.2014
// Version  : 1.1 (angepasst fuer uPlay)
// Autor    : UB
// EMail    : mc-4u(@)t-online.de
// Web      : www.mikrocontroller-4u.de
// CPU      : XMC1100-0064 (XMC2GO)
// IDE      : KEIL uVision 5.10
// ARMCC    : 5.04
// Module   : GPIO, XMC2GO_SPI
// Funktion : LCD-Display Nokia5110 (Chip=PCD8544)
//            Auflsung : 84 x 48 Pixel 
//            Schnittstelle = SPI (FRQ-Max = 4MHz)
//
// Hinweis  : Das Display benutzt die CPU-Pins :
//
//            P0.15 : Reset    (Display Pin 1 = RST)
//            P0.14 : D/C      (Display Pin 3 = DC)
//            P0.9  : CS       (Display Pin 2 = CE)
//            P0.8  : SPI_CLK  (Display Pin 5 = CLK)
//            P0.7  : SPI_MOSI (Display Pin 4 = DIN)
//            P0.6  : SPI_MISO (Break-Button)
//
//            Display Pin 6  = Vcc
//            Display Pin 8  = GND
//--------------------------------------------------------------



//--------------------------------------------------------------
// Includes
//--------------------------------------------------------------
#include "xmc2go_ub_nokia5110.h"


//--------------------------------------------------------------
uint8_t LCD_RAM[LCD_MAXX][LCD_LINES];
uint8_t lcd_inverse,lcd_collision;


//--------------------------------------------------------------
// interne Funktionen
//--------------------------------------------------------------
void P_NOKIA5110_InitIO(void);
void P_NOKIA5110_InitChip(void);
void P_NOKIA5110_RST(uint8_t temp);
void P_NOKIA5110_DC(uint8_t wert);
void P_NOKIA5110_Delay(volatile uint32_t nCount);
void P_NOKIA5110_CMD(uint8_t wert);
int16_t P_NOKIA5110_sgn(int16_t x);


//--------------------------------------------------------------
// Init vom LCD-Display
//--------------------------------------------------------------
void XMC2GO_LCD_Init(void)
{ 
  // init aller Variabeln
  lcd_inverse=0;
  lcd_collision=0;
  // init aller IO-Pins
  P_NOKIA5110_InitIO();
  // init vom SPI-BUS
  XMC2GO_Spi_Init();
  // LCD-Controller initialisieren
  P_NOKIA5110_InitChip();
}


//--------------------------------------------------------------
// Refresh vom Display Inhalt
// (muss zyklisch aufgerufen werden)
//
// (liesst auch den Status vom Break-Button aus)
//--------------------------------------------------------------
uint8_t XMC2GO_LCD_Refresh(void)
{
  uint8_t ret_wert=0;
  uint8_t x,y,wert;
	
  // Data
  P_NOKIA5110_DC(1);	
	
  for(y=0;y<LCD_LINES;y++) {
    for(x=0;x<LCD_MAXX;x++) {
      wert=LCD_RAM[x][y];
      // Wert senden/empfangen
      ret_wert=XMC2GO_Spi_SendByte(wert);						
    }
  }		
	
  return(ret_wert);
}


//--------------------------------------------------------------
// Inverse-Mode setzen
//--------------------------------------------------------------
void XMC2GO_LCD_Inverse(uint8_t mode)
{
  if(mode==0) {
    lcd_inverse=0;
  }		
  else {
    lcd_inverse=1;
  }
}


//--------------------------------------------------------------
// LCD-Display loeschen
//  inverse=0 -> alle Pixel loeschen
//  inverse=1 -> alle Pixel setzen
//--------------------------------------------------------------
void XMC2GO_LCD_Clear(void)
{
  uint8_t x,y;

  if(lcd_inverse==0) {
    for(y=0;y<LCD_LINES;y++) {
      for(x=0;x<LCD_MAXX;x++) {
        LCD_RAM[x][y]=0x00;
      }
    }		
  }
  else {
    for(y=0;y<LCD_LINES;y++) {
      for(x=0;x<LCD_MAXX;x++) {
        LCD_RAM[x][y]=0xFF;
      }
    }			
  }
}


//--------------------------------------------------------------
// setzt ein Pixel an x,y Position
//  inverse=0 -> Pixel setzen (check auf Kollision)
//  inverse=1 -> Pixel loeschen
//--------------------------------------------------------------
void XMC2GO_LCD_SetPixel(uint8_t xpos, uint8_t ypos)
{
  uint8_t y,wert;
  uint8_t lshift,maske=0x01;
	
  if(xpos>=LCD_MAXX) xpos=LCD_MAXX-1;
  if(ypos>=LCD_MAXY) xpos=LCD_MAXY-1;
	
  y=(ypos>>3);
  lshift=(ypos&0x07);
  maske=(maske<<lshift);
  wert=LCD_RAM[xpos][y];
  if(lcd_inverse==0) {
    if((wert&maske)!=0) lcd_collision=1;
    wert|=maske;
  }
  else {
    wert&=~maske;
  }
  LCD_RAM[xpos][y]=wert;
}


//--------------------------------------------------------------
// loescht ein Pixel an x,y Position
//  inverse=0 -> Pixel loeschen
//  inverse=1 -> Pixel setzen
//--------------------------------------------------------------
void XMC2GO_LCD_ClearPixel(uint8_t xpos, uint8_t ypos)
{
  uint8_t y,wert;
  uint8_t lshift,maske=0x01;
	
  if(xpos>=LCD_MAXX) xpos=LCD_MAXX-1;
  if(ypos>=LCD_MAXY) xpos=LCD_MAXY-1;
	
  y=(ypos>>3);
  lshift=(ypos&0x07);
  maske=(maske<<lshift);
  wert=LCD_RAM[xpos][y];
  if(lcd_inverse==0) {
    wert&=~maske;
  }
  else {
    wert|=maske;
  }
  LCD_RAM[xpos][y]=wert;
}


//--------------------------------------------------------------
// auslesen vom Pixel an x,y Position
// ret_wert : 
//    0 = Pixel ist geloescht
//    1 = Pixel ist gesetzt
//--------------------------------------------------------------
uint8_t XMC2GO_LCD_ReadPixel(uint8_t xpos, uint8_t ypos)
{
  uint8_t ret_wert=0;
  uint8_t y,wert;
  uint8_t lshift,maske=0x01;
	
  if(xpos>=LCD_MAXX) xpos=LCD_MAXX-1;
  if(ypos>=LCD_MAXY) xpos=LCD_MAXY-1;
	
  y=(ypos>>3);
  lshift=(ypos&0x07);
  maske=(maske<<lshift);
  wert=LCD_RAM[xpos][y];
  if((wert&maske)!=0) ret_wert=1;	
	
  return(ret_wert);
}


//--------------------------------------------------------------
// Zeichnet eine Linie
// von x1,y1 nach x2,y2
// (benutzt wird der Bresenham-Algorithmus)
//  inverse=0 -> schwarze linie
//  inverse=1 -> weisse linie
//--------------------------------------------------------------
void XMC2GO_LCD_DrawLine(uint8_t x1, uint8_t y1, uint8_t x2, uint8_t y2)
{
  int16_t x, y, t, dx, dy, incx, incy, pdx, pdy, ddx, ddy, es, el, err;
 
  // Entfernung in beiden Dimensionen berechnen
  dx = x2 - x1;
  dy = y2 - y1;
 
  // Vorzeichen des Inkrements bestimmen
  incx = P_NOKIA5110_sgn(dx);
  incy = P_NOKIA5110_sgn(dy);
  if(dx<0) dx = -dx;
  if(dy<0) dy = -dy;
 
  // feststellen, welche Entfernung grer ist
  if (dx>dy) {
    // x ist schnelle Richtung 
    pdx=incx; pdy=0;    // pd. ist Parallelschritt
    ddx=incx; ddy=incy; // dd. ist Diagonalschritt
    es =dy;   el =dx;   // Fehlerschritte schnell, langsam
  }
  else {
    // y ist schnelle Richtung
    pdx=0;    pdy=incy; // pd. ist Parallelschritt
    ddx=incx; ddy=incy; // dd. ist Diagonalschritt
    es =dx;   el =dy;   // Fehlerschritte schnell, langsam
  }
 
  // Initialisierungen vor Schleifenbeginn
  x = x1;
  y = y1;
  err = (el>>1);
  XMC2GO_LCD_SetPixel(x,y);
 
  // Pixel berechnen
  for(t=0; t<el; ++t) // t zaehlt die Pixel, el ist auch Anzahl
  {
    // Aktualisierung Fehlerterm
    err -= es; 
    if(err<0) {
      // Fehlerterm wieder positiv (>=0) machen
      err += el;
      // Schritt in langsame Richtung, Diagonalschritt
      x += ddx;
      y += ddy;
    }
    else {
      // Schritt in schnelle Richtung, Parallelschritt
      x += pdx;
      y += pdy;
    }
    XMC2GO_LCD_SetPixel(x,y);
  }
}


//--------------------------------------------------------------
// Zeichnet ein Rechteck
// von x1,y1 nach x2,y2
//  inverse=0 -> schwarzes rechteck
//  inverse=1 -> weisses rechteck
//--------------------------------------------------------------
void XMC2GO_LCD_DrawRect(uint8_t x1, uint8_t y1, uint8_t x2, uint8_t y2)
{
  XMC2GO_LCD_DrawLine(x1,y1,x2,y1);
  XMC2GO_LCD_DrawLine(x1,y2,x2,y2);
  XMC2GO_LCD_DrawLine(x1,y1,x1,y2);
  XMC2GO_LCD_DrawLine(x2,y1,x2,y2);
}


//--------------------------------------------------------------
// Fuellt einen Bereich
// von x1,y1 nach x2,y2
//  inverse=0 -> fuellt mit schwarz
//  inverse=1 -> fuellt mit weiss
//--------------------------------------------------------------
void XMC2GO_LCD_DrawFill(uint8_t x1, uint8_t y1, uint8_t x2, uint8_t y2)
{
  uint8_t n;
	
  if(x2>=x1) {
    for(n=x1;n<=x2;n++) {
      XMC2GO_LCD_DrawLine(n,y1,n,y2);			
    }
  }
  else {
    for(n=x2;n<=x1;n++) {
      XMC2GO_LCD_DrawLine(n,y1,n,y2);			
    }
  }
}


//--------------------------------------------------------------
// Zeichnet einen Kreis
// Mittelpunkt = x1,y1
// Radius = radius
// (benutzt wird der Bresenham-Algorithmus)
//  inverse=0 -> schwarzer kreis
//  inverse=1 -> weisser kreis
//--------------------------------------------------------------
void XMC2GO_LCD_DrawCircle(uint8_t x0, uint8_t y0, uint8_t radius)
{
  int16_t f=1-radius, ddF_x=0, ddF_y=-2*radius, x=0, y=radius;

  XMC2GO_LCD_SetPixel(x0, y0 + radius);
  XMC2GO_LCD_SetPixel(x0, y0 - radius);
  XMC2GO_LCD_SetPixel(x0 + radius, y0);
  XMC2GO_LCD_SetPixel(x0 - radius, y0);

  while(x < y)
  {
    if(f >= 0)
    {
      y--;
      ddF_y += 2;
      f += ddF_y;
    }
    x++;
    ddF_x += 2;
    f += ddF_x + 1;

    XMC2GO_LCD_SetPixel(x0 + x, y0 + y);
    XMC2GO_LCD_SetPixel(x0 - x, y0 + y);
    XMC2GO_LCD_SetPixel(x0 + x, y0 - y);
    XMC2GO_LCD_SetPixel(x0 - x, y0 - y);
    XMC2GO_LCD_SetPixel(x0 + y, y0 + x);
    XMC2GO_LCD_SetPixel(x0 - y, y0 + x);
    XMC2GO_LCD_SetPixel(x0 + y, y0 - x);
    XMC2GO_LCD_SetPixel(x0 - y, y0 - x);
  }
}


//--------------------------------------------------------------
// Zeichnet ein Ascii-Zeichen eines Fonts an x,y Position
// (Font = max 16 Pixel breite)
// -> Font muss mit &-Operator uebergeben werden
//  inverse=0 -> schwarze Schrift auf weissem Hintergrund
//  inverse=1 -> weisse Schrift auf schwarzem Hintergrund
//--------------------------------------------------------------
void XMC2GO_LCD_DrawChar(uint8_t x, uint8_t y, uint8_t ascii, UB_Font *font)
{
  uint16_t xn,yn,maske;
  const uint16_t *wert;
  
  ascii -= 32;
  wert=&font->table[ascii * font->height];

      
  for(yn = 0; yn < font->height; yn++) {
    maske=0x80;
    for(xn = 0; xn < font->width; xn++) {			
      if((wert[yn] & maske) == 0x00) {
        // Pixel loeschen
        XMC2GO_LCD_ClearPixel(xn+x,yn+y);
      }
      else {
        // Pixel setzen
        XMC2GO_LCD_SetPixel(xn+x,yn+y);
      }
      maske=(maske>>1);
    }
  }
}


//--------------------------------------------------------------
// Zeichnet einen String eines Fonts an x,y Position
// Font = max 16 Pixel breite)
// -> Font muss mit &-Operator uebergeben werden
//  inverse=0 -> schwarze Schrift auf weissem Hintergrund
//  inverse=1 -> weisse Schrift auf schwarzem Hintergrund
//--------------------------------------------------------------
void XMC2GO_LCD_DrawString(uint8_t x, uint8_t y,char *ptr, UB_Font *font)
{
  uint16_t pos;

  pos=x;
  while (*ptr != 0) {
    XMC2GO_LCD_DrawChar(pos,y,*ptr,font);
    pos+=font->width;
    ptr++;
  }
}

//--------------------------------------------------------------
// loeschen vom Kollisions wert
//--------------------------------------------------------------
void XMC2GO_LCD_ClearCollision(void)
{
  lcd_collision=0;
}

//--------------------------------------------------------------
// auslesen vom aktuellen Kollisions wert
// ret_wert : 
//    0 = es hat keine Kollision stattgefunden
//    1 = es hat eine Kollision stattgefunden
//--------------------------------------------------------------
uint8_t XMC2GO_LCD_GetCollision(void)
{
  uint8_t ret_wert=0;
  
  ret_wert=lcd_collision;
  lcd_collision=0;
  
  return(ret_wert);
}

//--------------------------------------------------------------
// interne Funktion
// Init aller IO-Pins fuer das Display
//--------------------------------------------------------------
void P_NOKIA5110_InitIO(void)
{	
  // DC-Pin
  P0_14_set_mode(OUTPUT_PP_GP);

  // Reset-Pin
  P0_15_set_mode(OUTPUT_PP_GP);
}


//--------------------------------------------------------------
// interne Funktion
// initialisiert den PCD8544-Controller im Display
//--------------------------------------------------------------
void P_NOKIA5110_InitChip(void)
{
  P_NOKIA5110_RST(0);
  P_NOKIA5110_DC(1);

  P_NOKIA5110_Delay(LCD_INIT_PAUSE);
  //Release Reset Pin
  P_NOKIA5110_RST(1);

  //Configure LCD module
  P_NOKIA5110_CMD(0x21);
  P_NOKIA5110_CMD(0xc0);
  P_NOKIA5110_CMD(0x14);
  P_NOKIA5110_CMD(0x06); 
  P_NOKIA5110_CMD(0x20); 
  P_NOKIA5110_CMD(0x0c);
  P_NOKIA5110_CMD(0x40);
  P_NOKIA5110_CMD(0x80);	
  XMC2GO_LCD_Clear(); 
}

//--------------------------------------------------------------
// interne Funktion
// Reset-Pin schalten
// wert [0=Lo, 1=Hi]
//--------------------------------------------------------------
void P_NOKIA5110_RST(uint8_t wert)
{
  if (wert == 0) {
    P0_15_reset();
  }
  else {
    P0_15_set();
  }
}


//--------------------------------------------------------------
// interne Funktion
// DC-Pin schalten
// wert [0=Lo, 1=Hi]
//--------------------------------------------------------------
void P_NOKIA5110_DC(uint8_t wert)
{
  if (wert == 0) {
    P0_14_reset();
  }
  else {
    P0_14_set();
  }
}


//--------------------------------------------------------------
// interne Funktion
// kleine Pause
//-------------------------------------------------------------- 
void P_NOKIA5110_Delay(volatile uint32_t nCount)
{
  volatile uint32_t index = 0;

  for(index = nCount; index != 0; index--);
}


//--------------------------------------------------------------
// interne Funktion
// Kommando per SPI an Display senden
//--------------------------------------------------------------
void P_NOKIA5110_CMD(uint8_t wert)
{
  // Kommando
  P_NOKIA5110_DC(0);

  // Wert senden
  XMC2GO_Spi_SendByte(wert);
}


//--------------------------------------------------------------
// interne Funktion
// Signum funktion
//  Return_Wert
//    1 ,wenn x > 0
//    0 ,wenn x = 0
//   -1 ,wenn x < 0
//--------------------------------------------------------------
int16_t P_NOKIA5110_sgn(int16_t x)
{
  return (x > 0) ? 1 : (x < 0) ? -1 : 0;
}

