/*
 * Copyright (c) 2006, Adam Dunkels
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */


//------------------------------------------------------------
// Changelog
//------------------------------------------------------------
//  [V:0.1 / 04.09.2013] :
// Author   : Adam Dunkels
// Web      : http://dunkels.com/adam/ubasic/
//
// original Version from Adam Dunkels without changings
//------------------------------------------------------------
//  [V:1.0 / 16.05.2014] :
// Author   : UB
// EMail    : mc-4u(@)t-online.de
// Web      : www.mikrocontroller-4u.de
//
// modified : removed all debug statements
// modified : replaced exit() statement with variable "failed"
// added    : function to read status of variable "failed"
// added    : function to read line-nr with failure
// bugfix   : variable "a" is permitted
// bugfix   : if/then/else support with :
//             "print","let","goto","gosub","for"
//             e.g.  "10 if i>3 then k=10 else k=20"
// modified : Token "then" not necessary
//             e.g.  "10 if i>3 k=10 else k=20"
// modified : negativ values possible
//             e.g   "10 a=-4"
//                   "20 b=-(c+3)"
// modified : removed unused Token "CALL"
// addded   : new Token "LINENUM" to find line numbers
// modified : variable type from char to integer
// modified : faster "jump_linenum" function
// added    : decrement support in "for/next"
//             e.g   "10 for i=10 to 2"
//                   "20 next i"
// added    : new Token "step" (if needed)
//             e.g   "10 for i=10 to 20 step 3"
//                   "20 next i"
// added    : with char ";" no newline after print
// added    : new Token "rem" to insert remark text
//             e.g.  "10 REM **** EXAMPLE-01 ****"
// added    : new Token for System-Function
//             "pause","clrtic","gettic"
// added    : new Tokens for Display-Functions
//             "inverse","cls","lcd","setpx","clrpx","getpx",
//             "line","rect","fill","circle","clrcoll","getcoll"
// added    : new Tokens for GPIO-Functions
//             "in","out"
// added    : new Token for ADC-Function
//             "adc"
// 
// added    : 3stage pre-parse function (see tokenizer)
//------------------------------------------------------------


#define  PLATFORM   0  //  0=XMC
//#define  PLATFORM   1  //  1=PC




#include "ubasic.h"
#include "tokenizer.h"

#include <stdio.h>
#include <stdlib.h>

#if PLATFORM==0
  #include "../ub_lib/xmc2go_ub_uart.h"
  #include "../ub_lib/xmc2go_ub_adc.h"
  #include "../ub_lib/xmc2go_ub_gpio.h"
  #include "../ub_lib/xmc2go_ub_nokia5110.h"
  #include "../ub_lib/xmc2go_ub_system.h"
#endif
#if PLATFORM==1
  #include "../xmc.h"
#endif

char *program_ptr;
#define MAX_STRINGLEN 50
static char string[MAX_STRINGLEN];
#if PLATFORM==0
static char tx_string[MAX_STRINGLEN];
#endif

#define MAX_GOSUB_STACK_DEPTH 10
static int gosub_stack[MAX_GOSUB_STACK_DEPTH];
static int gosub_stack_ptr;

struct for_state {
  int line_after_for;
  int for_variable;
  int to;
  int step;
};
#define MAX_FOR_STACK_DEPTH 4
static struct for_state for_stack[MAX_FOR_STACK_DEPTH];
static int for_stack_ptr;

#define MAX_VARNUM 26
static int variables[MAX_VARNUM];

static int ended;
static int failed, akt_linenum;  // added by UB

static int expr(void);
static void linenum_statement(void);
static void statement(void);
static void skipelse(void);  // added by UB





/*---------------------------------------------------------------------------*/
void
ubasic_init(char *program)
{
  int n;

  program_ptr = program;
  for_stack_ptr = gosub_stack_ptr = 0;
  ended = 0;
  failed =0;
  akt_linenum =0;
  // pre-parse the basic-program
  n=tokenizer_pre_parse(program);
  if(n!=0) {
    failed=UBASIC_PREPARSE_ERR;
    akt_linenum=tokenizer_linenum();
  }

  for(n=0;n<MAX_VARNUM;n++) {
    variables[n]=0;
  }
}
/*---------------------------------------------------------------------------*/
static void
accept(int token)
{
  if(token != tokenizer_token()) {
    tokenizer_error_print();
    failed=UBASIC_UNEXPECTED_TOKEN;
  }
  else {
    tokenizer_next();
  }
}
/*---------------------------------------------------------------------------*/
static int
varfactor(void)
{
  int r;
  r = ubasic_get_variable(tokenizer_variable_num());
  accept(TOKENIZER_VARIABLE);
  return r;
}
/*---------------------------------------------------------------------------*/
static int
factor(void)
{
  int r;

  switch(tokenizer_token()) {
  case TOKENIZER_NUMBER:
    r = tokenizer_num();
    accept(TOKENIZER_NUMBER);
    break;
  case TOKENIZER_LEFTPAREN:
    accept(TOKENIZER_LEFTPAREN);
    r = expr();
    accept(TOKENIZER_RIGHTPAREN);
    break;
  case TOKENIZER_MINUS:
    accept(TOKENIZER_MINUS);
    switch(tokenizer_token()) {
    case TOKENIZER_NUMBER:
      r = 0-tokenizer_num();
      accept(TOKENIZER_NUMBER);
      break;
    case TOKENIZER_LEFTPAREN:
      accept(TOKENIZER_LEFTPAREN);
      r = 0-expr();
      accept(TOKENIZER_RIGHTPAREN);
      break;
    default:
      r = 0-varfactor();
      break;
    }
    break;
  default:
    r = varfactor();
    break;
  }
  return r;
}
/*---------------------------------------------------------------------------*/
static int
term(void)
{
  int f1, f2;
  int op;

  f1 = factor();
  op = tokenizer_token();
  while(op == TOKENIZER_ASTR ||
	op == TOKENIZER_SLASH ||
	op == TOKENIZER_MOD) {
    tokenizer_next();
    f2 = factor();
    switch(op) {
    case TOKENIZER_ASTR:
      f1 = f1 * f2;
      break;
    case TOKENIZER_SLASH:
      f1 = f1 / f2;
      break;
    case TOKENIZER_MOD:
      f1 = f1 % f2;
      break;
    }
    op = tokenizer_token();
  }
  return f1;
}
/*---------------------------------------------------------------------------*/
static int
expr(void)
{
  int t1, t2;
  int op;

  t1 = term();
  op = tokenizer_token();
  while(op == TOKENIZER_PLUS ||
	op == TOKENIZER_MINUS ||
	op == TOKENIZER_AND ||
	op == TOKENIZER_OR) {
    tokenizer_next();
    t2 = term();
    switch(op) {
    case TOKENIZER_PLUS:
      t1 = t1 + t2;
      break;
    case TOKENIZER_MINUS:
      t1 = t1 - t2;
      break;
    case TOKENIZER_AND:
      t1 = t1 & t2;
      break;
    case TOKENIZER_OR:
      t1 = t1 | t2;
      break;
    }
    op = tokenizer_token();
  }
  return t1;
}
/*---------------------------------------------------------------------------*/
static int
relation(void)
{
  int r1, r2;
  int op;

  r1 = expr();
  op = tokenizer_token();
  while(op == TOKENIZER_LT ||
	op == TOKENIZER_GT ||
	op == TOKENIZER_EQ) {
    tokenizer_next();
    r2 = expr();
    switch(op) {
    case TOKENIZER_LT:
      r1 = r1 < r2;
      break;
    case TOKENIZER_GT:
      r1 = r1 > r2;
      break;
    case TOKENIZER_EQ:
      r1 = r1 == r2;
      break;
    }
    op = tokenizer_token();
  }
  return r1;
}
/*---------------------------------------------------------------------------*/
static void
jump_linenum(int linenum)
{
  int ok=0,akt_token,aktline;
	
  // try to jump from a pre-parsed linenumber
  ok=tokenizer_jumpfast(linenum);
  if(ok==1) return;

  // search the linenumber manually
  tokenizer_init(program_ptr);

  ok=0;
  do {
    akt_token=tokenizer_token();
    if(akt_token==TOKENIZER_ERROR) {
      ok=2;
    }
    else if(akt_token==TOKENIZER_ENDOFINPUT) {
      ok=1;
    }
    else if(akt_token==TOKENIZER_LINENUM) {
      aktline=tokenizer_num();
      if(aktline==linenum)  {
        ok=3;
      }
      else {
        tokenizer_skipline();
        tokenizer_next();
      }
    }
    else {
      ok=4;
    }
  }while(ok==0);
}
/*---------------------------------------------------------------------------*/
static void
goto_statement(void)
{
  accept(TOKENIZER_GOTO);
  jump_linenum(tokenizer_num());
}
/*---------------------------------------------------------------------------*/
#if PLATFORM==0 // XMC
static void
print_statement(void)
{
  int newline=0;

  accept(TOKENIZER_PRINT);
  do {
    if(tokenizer_token() == TOKENIZER_STRING) {
      tokenizer_string(string, sizeof(string));
      XMC2GO_Uart_SendString(string,NONE);
      tokenizer_next();
    } else if(tokenizer_token() == TOKENIZER_COMMA) {
      XMC2GO_Uart_SendString(" ",NONE);
      tokenizer_next();
    } else if(tokenizer_token() == TOKENIZER_SEMICOLON) {
      newline=1;
      tokenizer_next();
    } else if(tokenizer_token() == TOKENIZER_VARIABLE ||
	      tokenizer_token() == TOKENIZER_NUMBER ||
	      tokenizer_token() == TOKENIZER_MINUS) {
      sprintf(tx_string,"%d",expr());
      XMC2GO_Uart_SendString(tx_string,NONE);
    } else {
      break;
    }
  } while(tokenizer_token() != TOKENIZER_CR &&
	  tokenizer_token() != TOKENIZER_ENDOFINPUT);

  if(newline==0) XMC2GO_Uart_SendString("",CRLF);

  if(tokenizer_token()!=TOKENIZER_ELSE) {
    tokenizer_next();
  }
}
#endif
/*---------------------------------------------------------------------------*/
#if PLATFORM==1  // PC
static void
print_statement(void)
{
  int newline=0;

  accept(TOKENIZER_PRINT);
  do {
    if(tokenizer_token() == TOKENIZER_STRING) {
      tokenizer_string(string, sizeof(string));
      printf("%s", string);
      tokenizer_next();
    } else if(tokenizer_token() == TOKENIZER_COMMA) {
      printf(" ");
      tokenizer_next();
    } else if(tokenizer_token() == TOKENIZER_SEMICOLON) {
      newline=1;
      tokenizer_next();
    } else if(tokenizer_token() == TOKENIZER_VARIABLE ||
	      tokenizer_token() == TOKENIZER_NUMBER ||
	      tokenizer_token() == TOKENIZER_MINUS) {
      printf("%d", expr());
    } else {
      break;
    }
  } while(tokenizer_token() != TOKENIZER_CR &&
	  tokenizer_token() != TOKENIZER_ENDOFINPUT);

  if(newline==0) printf("\n");

  if(tokenizer_token()!=TOKENIZER_ELSE) {
    tokenizer_next();
  }
}
#endif
/*---------------------------------------------------------------------------*/
static void
if_statement(void)
{
  int r;

  accept(TOKENIZER_IF);

  r = relation();
  if(tokenizer_token()==TOKENIZER_THEN) {
    accept(TOKENIZER_THEN);
  }
  if(r) {
    statement();
    if(tokenizer_token()==TOKENIZER_ELSE) {
      skipelse();
    }
  }
  else {
    do {
      tokenizer_next();
    } while(tokenizer_token() != TOKENIZER_ELSE &&
	    tokenizer_token() != TOKENIZER_CR &&
	    tokenizer_token() != TOKENIZER_ENDOFINPUT);

    if(tokenizer_token() == TOKENIZER_ELSE) {
      tokenizer_next();
      statement();
    } else if(tokenizer_token() == TOKENIZER_CR) {
      tokenizer_next();
    }
  }
}
/*---------------------------------------------------------------------------*/
static void
let_statement(void)
{
  int var;

  var = tokenizer_variable_num();

  accept(TOKENIZER_VARIABLE);
  accept(TOKENIZER_EQ);
  ubasic_set_variable(var, expr());
  if(tokenizer_token()!=TOKENIZER_ELSE) {
    accept(TOKENIZER_CR);
  }
}
/*---------------------------------------------------------------------------*/
static void
gosub_statement(void)
{
  int linenum;
  accept(TOKENIZER_GOSUB);
  linenum = tokenizer_num();
  accept(TOKENIZER_NUMBER);
  if(tokenizer_token()!=TOKENIZER_ELSE) {
    accept(TOKENIZER_CR);
  }
  else {
    skipelse();
  }

  if(gosub_stack_ptr < MAX_GOSUB_STACK_DEPTH) {
    gosub_stack[gosub_stack_ptr] = tokenizer_num();
    gosub_stack_ptr++;
    jump_linenum(linenum);
  }
}
/*---------------------------------------------------------------------------*/
static void
return_statement(void)
{
  accept(TOKENIZER_RETURN);
  if(gosub_stack_ptr > 0) {
    gosub_stack_ptr--;
    jump_linenum(gosub_stack[gosub_stack_ptr]);
  }
}
/*---------------------------------------------------------------------------*/
static void
next_statement(void)
{
  int var,ist,soll,step;

  accept(TOKENIZER_NEXT);
  var = tokenizer_variable_num();
  accept(TOKENIZER_VARIABLE);
  if(for_stack_ptr > 0 &&
    var == for_stack[for_stack_ptr - 1].for_variable) {

    ist=ubasic_get_variable(var);
    soll=for_stack[for_stack_ptr - 1].to;
    step=for_stack[for_stack_ptr - 1].step;

    if(ist<soll) {
      // count up
      ubasic_set_variable(var,ist + step);
      if(ubasic_get_variable(var) <= for_stack[for_stack_ptr - 1].to) {
        jump_linenum(for_stack[for_stack_ptr - 1].line_after_for);
      } else {
        for_stack_ptr--;
        accept(TOKENIZER_CR);
      }
    }
    else if(ist>soll) {
      // count down
      ubasic_set_variable(var,ist - step);
      if(ubasic_get_variable(var) >= for_stack[for_stack_ptr - 1].to) {
        jump_linenum(for_stack[for_stack_ptr - 1].line_after_for);
      } else {
        for_stack_ptr--;
        accept(TOKENIZER_CR);
      }
    }
    else {
      for_stack_ptr--;
      accept(TOKENIZER_CR);
    }

  } else {
    accept(TOKENIZER_CR);
  }

}
/*---------------------------------------------------------------------------*/
static void
for_statement(void)
{
  int for_variable, to, step=1;

  accept(TOKENIZER_FOR);
  for_variable = tokenizer_variable_num();
  accept(TOKENIZER_VARIABLE);
  accept(TOKENIZER_EQ);
  ubasic_set_variable(for_variable, expr());
  accept(TOKENIZER_TO);
  to = expr();
  if(tokenizer_token()!=TOKENIZER_ELSE) {
    if(tokenizer_token()==TOKENIZER_STEP) {
      accept(TOKENIZER_STEP);
      step = expr();
      if(step<=0) step=1;
      if(tokenizer_token()!=TOKENIZER_ELSE) {
        accept(TOKENIZER_CR);
      }
      else {
        skipelse();
      }
    }
    else {
      accept(TOKENIZER_CR);
    }
  }
  else {
    skipelse();
  }

  if(for_stack_ptr < MAX_FOR_STACK_DEPTH) {
    for_stack[for_stack_ptr].line_after_for = tokenizer_num();
    for_stack[for_stack_ptr].for_variable = for_variable;
    for_stack[for_stack_ptr].to = to;
    for_stack[for_stack_ptr].step = step;

    for_stack_ptr++;
  }
}
/*---------------------------------------------------------------------------*/
static void
end_statement(void)
{
  accept(TOKENIZER_END);
  ended = 1;
}


/*---------------------------------------------------------------------------*/
// NEW STATEMENTS BY UB
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
static void
rem_statement(void)
{
  tokenizer_skipline();
  accept(TOKENIZER_CR);
}
/*---------------------------------------------------------------------------*/
static void
pause_statement(void)
{
  int ms=0;

  accept(TOKENIZER_PAUSE);
  if((tokenizer_token() == TOKENIZER_VARIABLE) ||
     (tokenizer_token() == TOKENIZER_NUMBER)) {
    ms=expr();
    if(ms<0) ms=0;
  }

  XMC2GO_Systic_Pause_ms(ms);
  if(tokenizer_token()!=TOKENIZER_ELSE) {
    accept(TOKENIZER_CR);
  }
  else {
    skipelse();
  }
}
/*---------------------------------------------------------------------------*/
static void
clrtic_statement(void)
{
  accept(TOKENIZER_CLRTIC);

  XMC2GO_Systic_ClrTic();
  if(tokenizer_token()!=TOKENIZER_ELSE) {
    accept(TOKENIZER_CR);
  }
  else {
    skipelse();
  }
}
/*---------------------------------------------------------------------------*/
static void
gettic_statement(void)
{
  int var=0,v=0;

  accept(TOKENIZER_GETTIC);
  var = tokenizer_variable_num();
  accept(TOKENIZER_VARIABLE);
  v=XMC2GO_Systic_GetTic();
  ubasic_set_variable(var, v);
  if(tokenizer_token()!=TOKENIZER_ELSE) {
    accept(TOKENIZER_CR);
  }
  else {
    skipelse();
  }
}
/*---------------------------------------------------------------------------*/
static void
inverse_statement(void)
{
  int v=0;

  accept(TOKENIZER_INVERSE);
  if((tokenizer_token() == TOKENIZER_VARIABLE) ||
     (tokenizer_token() == TOKENIZER_NUMBER)) {
    v=expr();
    if(v<0) v=0;
  }

  XMC2GO_LCD_Inverse(v);
  if(tokenizer_token()!=TOKENIZER_ELSE) {
    accept(TOKENIZER_CR);
  }
  else {
    skipelse();
  }
}
/*---------------------------------------------------------------------------*/
static void
cls_statement(void)
{
  accept(TOKENIZER_CLS);
  XMC2GO_LCD_Clear();
  if(tokenizer_token()!=TOKENIZER_ELSE) {
    accept(TOKENIZER_CR);
  }
  else {
    skipelse();
  }
}
/*---------------------------------------------------------------------------*/
static void
printlcd_statement(void)
{
  int x=0,y=0;

  accept(TOKENIZER_PRINT_LCD);
  if((tokenizer_token() == TOKENIZER_VARIABLE) ||
     (tokenizer_token() == TOKENIZER_NUMBER)) {
    x=expr();
    if(x<0) x=0;
  }
  accept(TOKENIZER_COMMA);
  if((tokenizer_token() == TOKENIZER_VARIABLE) ||
     (tokenizer_token() == TOKENIZER_NUMBER)) {
    y=expr();
    if(y<0) y=0;
  }
  accept(TOKENIZER_COMMA);
  if(tokenizer_token() == TOKENIZER_STRING) {
    tokenizer_string(string, sizeof(string));
    tokenizer_next();
  }
  else if(tokenizer_token() == TOKENIZER_VARIABLE ||
          tokenizer_token() == TOKENIZER_NUMBER ||
          tokenizer_token() == TOKENIZER_MINUS) {
    sprintf(string,"%d",expr());
  }

  XMC2GO_LCD_DrawString(x,y,string,&Font_6x8);
  if(tokenizer_token()!=TOKENIZER_ELSE) {
    accept(TOKENIZER_CR);
  }
  else {
    skipelse();
  }
}
/*---------------------------------------------------------------------------*/
static void
setpixel_statement(void)
{
  int x=0,y=0;

  accept(TOKENIZER_SETPIXEL);
  if((tokenizer_token() == TOKENIZER_VARIABLE) ||
     (tokenizer_token() == TOKENIZER_NUMBER)) {
    x=expr();
    if(x<0) x=0;
  }
  accept(TOKENIZER_COMMA);
  if((tokenizer_token() == TOKENIZER_VARIABLE) ||
     (tokenizer_token() == TOKENIZER_NUMBER)) {
    y=expr();
    if(y<0) y=0;
  }

  XMC2GO_LCD_SetPixel(x,y);
  if(tokenizer_token()!=TOKENIZER_ELSE) {
    accept(TOKENIZER_CR);
  }
  else {
    skipelse();
  }
}
/*---------------------------------------------------------------------------*/
static void
clrpixel_statement(void)
{
  int x=0,y=0;

  accept(TOKENIZER_CLRPIXEL);
  if((tokenizer_token() == TOKENIZER_VARIABLE) ||
     (tokenizer_token() == TOKENIZER_NUMBER)) {
    x=expr();
    if(x<0) x=0;
  }
  accept(TOKENIZER_COMMA);
  if((tokenizer_token() == TOKENIZER_VARIABLE) ||
     (tokenizer_token() == TOKENIZER_NUMBER)) {
    y=expr();
    if(y<0) y=0;
  }

  XMC2GO_LCD_ClearPixel(x,y);
  if(tokenizer_token()!=TOKENIZER_ELSE) {
    accept(TOKENIZER_CR);
  }
  else {
    skipelse();
  }
}
/*---------------------------------------------------------------------------*/
static void
getpixel_statement(void)
{
  int x=0,y=0,v=0,var=0;
	
  accept(TOKENIZER_GETPIXEL);
  var = tokenizer_variable_num();	
  accept(TOKENIZER_VARIABLE);
  accept(TOKENIZER_EQ);
  if((tokenizer_token() == TOKENIZER_VARIABLE) ||
     (tokenizer_token() == TOKENIZER_NUMBER)) {
    x=expr();
    if(x<0) x=0;
  }	
  accept(TOKENIZER_COMMA);
  if((tokenizer_token() == TOKENIZER_VARIABLE) ||
     (tokenizer_token() == TOKENIZER_NUMBER)) {
    y=expr();
    if(y<0) y=0;
  }		
  v=XMC2GO_LCD_ReadPixel(x,y);
  ubasic_set_variable(var, v);
  if(tokenizer_token()!=TOKENIZER_ELSE) {
    accept(TOKENIZER_CR);
  }
  else {
    skipelse();
  }
}
/*---------------------------------------------------------------------------*/
static void
line_statement(void)
{
  int x1=0,y1=0,x2=0,y2=0;

  accept(TOKENIZER_LINE);
  if((tokenizer_token() == TOKENIZER_VARIABLE) ||
     (tokenizer_token() == TOKENIZER_NUMBER)) {
    x1=expr();
    if(x1<0) x1=0;
  }
  accept(TOKENIZER_COMMA);
  if((tokenizer_token() == TOKENIZER_VARIABLE) ||
     (tokenizer_token() == TOKENIZER_NUMBER)) {
    y1=expr();
    if(y1<0) y1=0;
  }
  accept(TOKENIZER_COMMA);
  if((tokenizer_token() == TOKENIZER_VARIABLE) ||
     (tokenizer_token() == TOKENIZER_NUMBER)) {
    x2=expr();
    if(x2<0) x2=0;
  }
  accept(TOKENIZER_COMMA);
  if((tokenizer_token() == TOKENIZER_VARIABLE) ||
     (tokenizer_token() == TOKENIZER_NUMBER)) {
    y2=expr();
    if(y2<0) y2=0;
  }

  XMC2GO_LCD_DrawLine(x1,y1,x2,y2);
  if(tokenizer_token()!=TOKENIZER_ELSE) {
    accept(TOKENIZER_CR);
  }
  else {
    skipelse();
  }
}
/*---------------------------------------------------------------------------*/
static void
rect_statement(void)
{
  int x1=0,y1=0,x2=0,y2=0;

  accept(TOKENIZER_RECT);
  if((tokenizer_token() == TOKENIZER_VARIABLE) ||
     (tokenizer_token() == TOKENIZER_NUMBER)) {
    x1=expr();
    if(x1<0) x1=0;
  }
  accept(TOKENIZER_COMMA);
  if((tokenizer_token() == TOKENIZER_VARIABLE) ||
     (tokenizer_token() == TOKENIZER_NUMBER)) {
    y1=expr();
    if(y1<0) y1=0;
  }
  accept(TOKENIZER_COMMA);
  if((tokenizer_token() == TOKENIZER_VARIABLE) ||
     (tokenizer_token() == TOKENIZER_NUMBER)) {
    x2=expr();
    if(x2<0) x2=0;
  }
  accept(TOKENIZER_COMMA);
  if((tokenizer_token() == TOKENIZER_VARIABLE) ||
     (tokenizer_token() == TOKENIZER_NUMBER)) {
    y2=expr();
    if(y2<0) y2=0;
  }

  XMC2GO_LCD_DrawRect(x1,y1,x2,y2);
  if(tokenizer_token()!=TOKENIZER_ELSE) {
    accept(TOKENIZER_CR);
  }
  else {
    skipelse();
  }
}
/*---------------------------------------------------------------------------*/
static void
fill_statement(void)
{
  int x1=0,y1=0,x2=0,y2=0;

  accept(TOKENIZER_FILL);
  if((tokenizer_token() == TOKENIZER_VARIABLE) ||
     (tokenizer_token() == TOKENIZER_NUMBER)) {
    x1=expr();
    if(x1<0) x1=0;
  }
  accept(TOKENIZER_COMMA);
  if((tokenizer_token() == TOKENIZER_VARIABLE) ||
     (tokenizer_token() == TOKENIZER_NUMBER)) {
    y1=expr();
    if(y1<0) y1=0;
  }
  accept(TOKENIZER_COMMA);
  if((tokenizer_token() == TOKENIZER_VARIABLE) ||
     (tokenizer_token() == TOKENIZER_NUMBER)) {
    x2=expr();
    if(x2<0) x2=0;
  }
  accept(TOKENIZER_COMMA);
  if((tokenizer_token() == TOKENIZER_VARIABLE) ||
     (tokenizer_token() == TOKENIZER_NUMBER)) {
    y2=expr();
    if(y2<0) y2=0;
  }

  XMC2GO_LCD_DrawFill(x1,y1,x2,y2);
  if(tokenizer_token()!=TOKENIZER_ELSE) {
    accept(TOKENIZER_CR);
  }
  else {
    skipelse();
  }
}
/*---------------------------------------------------------------------------*/
static void
circle_statement(void)
{
  int x=0,y=0,r=0;

  accept(TOKENIZER_CIRCLE);
  if((tokenizer_token() == TOKENIZER_VARIABLE) ||
     (tokenizer_token() == TOKENIZER_NUMBER)) {
    x=expr();
    if(x<0) x=0;
  }
  accept(TOKENIZER_COMMA);
  if((tokenizer_token() == TOKENIZER_VARIABLE) ||
     (tokenizer_token() == TOKENIZER_NUMBER)) {
    y=expr();
    if(y<0) y=0;
  }
  accept(TOKENIZER_COMMA);
  if((tokenizer_token() == TOKENIZER_VARIABLE) ||
     (tokenizer_token() == TOKENIZER_NUMBER)) {
    r=expr();
    if(r<0) r=0;
  }

  XMC2GO_LCD_DrawCircle(x,y,r);
  if(tokenizer_token()!=TOKENIZER_ELSE) {
    accept(TOKENIZER_CR);
  }
  else {
    skipelse();
  }
}
/*---------------------------------------------------------------------------*/
static void
clrcollision_statement(void)
{
  accept(TOKENIZER_CLRCOLL);
  XMC2GO_LCD_ClearCollision();
  if(tokenizer_token()!=TOKENIZER_ELSE) {
    accept(TOKENIZER_CR);
  }
  else {
    skipelse();
  }
}
/*---------------------------------------------------------------------------*/
static void
getcollision_statement(void)
{
  int var=0,v=0;

  accept(TOKENIZER_GETCOLL);
  var = tokenizer_variable_num();
  accept(TOKENIZER_VARIABLE);
  v=XMC2GO_LCD_GetCollision();
  ubasic_set_variable(var, v);
  if(tokenizer_token()!=TOKENIZER_ELSE) {
    accept(TOKENIZER_CR);
  }
  else {
    skipelse();
  }
}
/*---------------------------------------------------------------------------*/
static void
in_statement(void)
{
  int var=0,p=0,v=0;

  accept(TOKENIZER_IN);
  var = tokenizer_variable_num();
  accept(TOKENIZER_VARIABLE);
  accept(TOKENIZER_EQ);
  if((tokenizer_token() == TOKENIZER_VARIABLE) ||
     (tokenizer_token() == TOKENIZER_NUMBER)) {
    p=expr();
    if(p<0) p=0;
  }
  v=XMC2GO_DIn_Read(p);
  ubasic_set_variable(var, v);
  if(tokenizer_token()!=TOKENIZER_ELSE) {
    accept(TOKENIZER_CR);
  }
  else {
    skipelse();
  }
}
/*---------------------------------------------------------------------------*/
static void
out_statement(void)
{
  int p=0,v=0;

  accept(TOKENIZER_OUT);
  if((tokenizer_token() == TOKENIZER_VARIABLE) ||
     (tokenizer_token() == TOKENIZER_NUMBER)) {
    p=expr();
    if(p<0) p=0;
  }
  accept(TOKENIZER_COMMA);
  if((tokenizer_token() == TOKENIZER_VARIABLE) ||
     (tokenizer_token() == TOKENIZER_NUMBER)) {
    v=expr();
    if(v<0) v=0;
  }

  if(v==0) {
    XMC2GO_DOut_Lo(p);
  }
  else if(v==1) {
    XMC2GO_DOut_Hi(p);
  }
  else if(v==2) {
    XMC2GO_DOut_Toggle(p);
  }
  if(tokenizer_token()!=TOKENIZER_ELSE) {
    accept(TOKENIZER_CR);
  }
  else {
    skipelse();
  }
}
/*---------------------------------------------------------------------------*/
static void
adc_statement(void)
{
  int var=0,p=0,v;

  accept(TOKENIZER_ADC);
  var = tokenizer_variable_num();
  accept(TOKENIZER_VARIABLE);
  accept(TOKENIZER_EQ);
  if((tokenizer_token() == TOKENIZER_VARIABLE) ||
     (tokenizer_token() == TOKENIZER_NUMBER)) {
    p=expr();
    if(p<0) p=0;
  }
  v=XMC2GO_Adc_Read(p);
  ubasic_set_variable(var, v);
  if(tokenizer_token()!=TOKENIZER_ELSE) {
    accept(TOKENIZER_CR);
  }
  else {
    skipelse();
  }
}

/*---------------------------------------------------------------------------*/
// END OF NEW STATEMENTS BY UB
/*---------------------------------------------------------------------------*/




/*---------------------------------------------------------------------------*/
static void
statement(void)
{
  int token;

  if(failed!=0) return;

  token = tokenizer_token();

  switch(token) {
  case TOKENIZER_PRINT:
    print_statement();
    break;
  case TOKENIZER_IF:
    if_statement();
    break;
  case TOKENIZER_GOTO:
    goto_statement();
    break;
  case TOKENIZER_GOSUB:
    gosub_statement();
    break;
  case TOKENIZER_RETURN:
    return_statement();
    break;
  case TOKENIZER_FOR:
    for_statement();
    break;
  case TOKENIZER_NEXT:
    next_statement();
    break;
  case TOKENIZER_END:
    end_statement();
    break;
  case TOKENIZER_LET:
    accept(TOKENIZER_LET);
    /* Fall through. */
  case TOKENIZER_VARIABLE:
    let_statement();
    break;
  //----------------------------------------
  // NEW TOKENS BY UB
  //----------------------------------------
  case TOKENIZER_REM :
    rem_statement();
    break;
  case TOKENIZER_PAUSE :
    pause_statement();
    break;
  case TOKENIZER_CLRTIC :
    clrtic_statement();
    break;
  case TOKENIZER_GETTIC :
    gettic_statement();
    break;
  case TOKENIZER_INVERSE :
    inverse_statement();
    break;	
  case TOKENIZER_CLS :
    cls_statement();
    break;
  case TOKENIZER_PRINT_LCD :
    printlcd_statement();
    break;
  case TOKENIZER_SETPIXEL :
    setpixel_statement();
    break;
  case TOKENIZER_CLRPIXEL :
    clrpixel_statement();
    break;	
  case TOKENIZER_GETPIXEL :
    getpixel_statement();
    break;	
  case TOKENIZER_LINE :
    line_statement();
    break;
  case TOKENIZER_RECT :
    rect_statement();
    break;
  case TOKENIZER_FILL :
    fill_statement();
    break;	
  case TOKENIZER_CIRCLE :
    circle_statement();
    break;
  case TOKENIZER_CLRCOLL :
    clrcollision_statement();
    break;   
  case TOKENIZER_GETCOLL :
    getcollision_statement();
    break;  
  case TOKENIZER_IN :
    in_statement();
    break;
  case TOKENIZER_OUT :
    out_statement();
    break;
  case TOKENIZER_ADC :
    adc_statement();
    break;
  //----------------------------------------
  default:
    failed=UBASIC_INVALID_KEYWORD;
  }
}
/*---------------------------------------------------------------------------*/
static void
linenum_statement(void)
{
  if(failed==0) akt_linenum = tokenizer_num();
  accept(TOKENIZER_LINENUM);
  statement();
  return;
}
/*---------------------------------------------------------------------------*/
void
ubasic_run(void)
{
  if(tokenizer_finished()) {
    return;
  }

  linenum_statement();
}
/*---------------------------------------------------------------------------*/
int
ubasic_finished(void)
{
  return ended || tokenizer_finished() || failed;
}
/*---------------------------------------------------------------------------*/
void
ubasic_set_variable(int varnum, int value)
{
  if(varnum >= 0 && varnum <= MAX_VARNUM) {
    variables[varnum] = value;
  }
}
/*---------------------------------------------------------------------------*/
int
ubasic_get_variable(int varnum)
{
  if(varnum >= 0 && varnum <= MAX_VARNUM) {
    return variables[varnum];
  }
  return 0;
}
/*---------------------------------------------------------------------------*/



/*---------------------------------------------------------------------------*/
// NEW FUNCTIONS BY UB
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
static
void skipelse(void)
{
  do {
    tokenizer_next();
  } while(tokenizer_token() != TOKENIZER_CR &&
          tokenizer_token() != TOKENIZER_ENDOFINPUT);

  if(tokenizer_token() == TOKENIZER_CR) {
    accept(TOKENIZER_CR);
  }
}
/*---------------------------------------------------------------------------*/
int
ubasic_failed(void)
{
  return failed;
}
/*---------------------------------------------------------------------------*/
int
ubasic_linenum(void)
{
  return akt_linenum;
}
/*---------------------------------------------------------------------------*/
